@@ -10,6 +10,9 @@ import re
 from errno import EIO, ENOENT
 from stat import S_IFDIR, S_IFREG
 from threading import Timer
+from tenacity import retry, wait_exponential
+
+from time import sleep
 from time import time
 
 import boto3
@@ -115,9 +118,10 @@ class FtpFetcher:
 class HttpFetcher:
     SSL_VERIFY = os.environ.get("SSL_VERIFY", True) not in FALSY
 
-    def __init__(self):
+    def __init__(self, logger):
+        self.logger = logger
         if not self.SSL_VERIFY:
-            logging.warning(
+            logger.warning(
                 "You have set ssl certificates to not be verified. "
                 "This may leave you vulnerable. "
                 "http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification"
@@ -139,15 +143,21 @@ class HttpFetcher:
             if match:
                 return int(match.group(1))
 
-            logging.error(traceback.format_exc())
+            self.logger.error(traceback.format_exc())
             raise FuseOSError(ENOENT)
 
+    @retry(wait=wait_exponential(multiplier=1, min=4, max=10))
     def get_data(self, url, start, end):
         headers = {"Range": "bytes={}-{}".format(start, end), "Accept-Encoding": ""}
+        self.logger.info("gettings %s %s %s", url, start, end)
         r = requests.get(url, headers=headers)
+        self.logger.info("got %s", r.status_code)
+        print("got", r.status_code)
+        r.raise_for_status()
         block_data = np.frombuffer(r.content, dtype=np.uint8)
         return block_data
 
+
 class S3Fetcher:
     SSL_VERIFY = os.environ.get("SSL_VERIFY", True) not in FALSY
 
@@ -155,30 +165,35 @@ class S3Fetcher:
         self.logger = logger
         self.logger.info("Creating S3Fetcher with aws_profile=%s", aws_profile)
         self.session = boto3.Session(profile_name=aws_profile)
-        self.client = self.session.client('s3')
+        self.client = self.session.client("s3")
         pass
 
     def parse_bucket_key(self, url):
         url_parts = urlparse(url, allow_fragments=False)
         bucket = url_parts.netloc
-        key = url_parts.path.strip('/')
+        key = url_parts.path.strip("/")
 
         return bucket, key
+
     def get_size(self, url):
         bucket, key = self.parse_bucket_key(url)
 
         response = self.client.head_object(Bucket=bucket, Key=key)
-        size = response['ContentLength']
+        size = response["ContentLength"]
         return size
 
+    @retry(wait=wait_exponential(multiplier=1, min=4, max=10))
     def get_data(self, url, start, end):
         bucket, key = self.parse_bucket_key(url)
-        obj = boto3.resource('s3').Object(bucket, key)
-        stream = self.client.get_object(Bucket=bucket, Key=key, Range="bytes={}-{}".format(start, end))['Body']
+        obj = boto3.resource("s3").Object(bucket, key)
+        stream = self.client.get_object(
+            Bucket=bucket, Key=key, Range="bytes={}-{}".format(start, end)
+        )["Body"]
         contents = stream.read()
         block_data = np.frombuffer(contents, dtype=np.uint8)
         return block_data
 
+
 class HttpFs(LoggingMixIn, Operations):
     """
     A read only http/https/ftp filesystem.
@@ -201,15 +216,16 @@ class HttpFs(LoggingMixIn, Operations):
         self.logger = logger
         self.last_report_time = 0
         self.total_requests = 0
+        self.getting = set()
 
         if not self.logger:
             self.logger = logging.getLogger(__name__)
 
         if schema == "http" or schema == "https":
-            self.fetcher = HttpFetcher()
+            self.fetcher = HttpFetcher(self.logger)
         elif schema == "ftp":
             self.fetcher = FtpFetcher()
-        elif schema == 's3':
+        elif schema == "s3":
             self.fetcher = S3Fetcher(aws_profile, self.logger)
         else:
             raise ("Unknown schema: {}".format(schema))
@@ -240,13 +256,17 @@ class HttpFs(LoggingMixIn, Operations):
                 self.lru_attrs[path] = dict(st_mode=(S_IFDIR | 0o555), st_nlink=2)
                 return self.lru_attrs[path]
 
-            if path[-2:] != ".." and not path.endswith('..-journal') and not path.endswith('..-wal'):
+            if (
+                path[-2:] != ".."
+                and not path.endswith("..-journal")
+                and not path.endswith("..-wal")
+            ):
                 return dict(st_mode=(S_IFDIR | 0o555), st_nlink=2)
 
             url = "{}:/{}".format(self.schema, path[:-2])
 
             # there's an exception for the -jounral files created by SQLite
-            if not path.endswith('..-journal') and not path.endswith('..-wal'):
+            if not path.endswith("..-journal") and not path.endswith("..-wal"):
                 size = self.getSize(url)
             else:
                 size = 0
@@ -284,13 +304,21 @@ class HttpFs(LoggingMixIn, Operations):
     def read(self, path, size, offset, fh):
         t1 = time()
 
+        # print("read %s %s %s", path, offset, size)
+
         if t1 - self.last_report_time > REPORT_INTERVAL:
+            """
             self.logger.info(
-                        "lru hits: {} lru misses: {} disk hits: {} total_requests: {}".format(
-                            self.lru_hits, self.lru_misses, self.disk_hits, self.disk_misses, self.total_requests
-                        )
-                    )
-            self.last_report_time = t1
+                "lru hits: {} lru misses: {} disk hits: {} total_requests: {}".format(
+                    self.lru_hits,
+                    self.lru_misses,
+                    self.disk_hits,
+                    self.disk_misses,
+                    self.total_requests,
+                )
+            )
+            """
+            pass
         try:
             self.total_requests += 1
             if path in self.lru_attrs:
@@ -299,7 +327,10 @@ class HttpFs(LoggingMixIn, Operations):
                 self.logger.debug("read url: {}".format(url))
                 self.logger.debug(
                     "offset: {} - {} request_size (KB): {:.2f} block: {}".format(
-                        offset, offset + size - 1, size / 2 ** 10, offset // self.block_size
+                        offset,
+                        offset + size - 1,
+                        size / 2 ** 10,
+                        offset // self.block_size,
                     )
                 )
                 output = np.zeros((size,), np.uint8)
@@ -314,7 +345,13 @@ class HttpFs(LoggingMixIn, Operations):
                     block_num = curr_start // self.block_size
                     block_start = self.block_size * (curr_start // self.block_size)
 
+                    block_id = (url, block_num)
+                    while block_id in self.getting:
+                        sleep(0.05)
+
+                    self.getting.add(block_id)
                     block_data = self.get_block(url, block_num)
+                    self.getting.remove(block_id)
 
                     data_start = (
                         curr_start - (curr_start // self.block_size) * self.block_size
@@ -378,6 +415,7 @@ class HttpFs(LoggingMixIn, Operations):
             self.disk_misses += 1
             block_start = block_num * self.block_size
 
+            print("getting data", url, block_start, block_start + self.block_size - 1)
             block_data = self.fetcher.get_data(
                 url, block_start, block_start + self.block_size - 1
             )
